//////////////////////////////////////////////////////////////
//
//  Whupsey - A Whups Client Library in C
//
//  21.11.2002
//

#include <stdlib.h>
#include <stdio.h>
#include "whupsey.h"

static char *statec[] = {"Unknown", "Unconfirmed", "New", "Assigned", "Resolved"};

void test_summary_ids     (Whupsey *w);
void test_create_ticket   (Whupsey *w);
void test_display_types   (Whupsey *w);
void test_display_modules (Whupsey *w);
void test_set_attributes  (Whupsey *w, unsigned int ticket_id, WhupseyType *type);

static char *server;
static char *user;
static char *password;
static int   create = 0;

void usage()
{
    printf("Usage: whupseytest [-c] server user password\n");
    printf("   eg. whupseytest http://example.com/horde/rpc.php jrandom password\n\n");
    printf("   '-c' will attempt to create a ticket in the remote whups server\n");
    printf("   otherwise, it will print the modules, types, states etc.\n\n");
    exit(1); 
}

void getargs(int argc, char **argv)
{
    int s = 0;

    if (argc > 1 && !strcmp(argv[1], "-c"))
        s = create = 1;

    if (argc < 3 + s)
    {
        server   = getenv("HORDE_SERVER");
        user     = getenv("HORDE_USER");
        password = getenv("HORDE_PASSWORD");

        if (!server || !user || !password)
            usage();
    }
    else
    {
        server   = argv[1 + s];
        user     = argv[2 + s];
        password = argv[3 + s];
    }
}

int main(int argc, char **argv)
{
    WhupseyResult  r;
    Whupsey       *w;

    getargs(argc, argv);

    printf("Attempting to connect to whups server:\n");
    printf("   Server: %s\n", server);
    printf("   User:   %s\n", user);
    printf("\n");

    whupsey_create(&w, "WhupseyTest", "1.0", server, user, password);

    r = whupsey_load_schema(w);
    if (r != WHUPSEY_OK)
    {
        if (r == WHUPSEY_RPCFAULT)
            printf("Error: Couldn't load schema: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
        else if (r == WHUPSEY_NOMEMORY)
            printf("Error: Couldn't load schema: Out of memory\n");
        else
            printf("Error: Couldn't load schema: Unknown error\n");

        exit(1);
    }

    test_display_types(w);
    test_display_modules(w);

    if (create)
        test_create_ticket(w);

    whupsey_free(w);
    return 0;
}

void display_ids(Whupsey *w, const char *title, int *ids, int c)
{
    int i;

    printf("%s:\n", title);

    if (c == 0)
    {
        printf("    None\n");
    }
    else
    {
        for (i = 0; i < c; i++)
            printf("    %d\n", ids[i]);
    }

    printf("\n");
}

void test_summary_ids(Whupsey *w)
{
    int *ids;
    int  c;

    if (whupsey_get_assigned_tickets(w, &ids, &c) == WHUPSEY_RPCFAULT)
    {
        printf("Error: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    }
    else
    {
        display_ids(w, "Assigned Tickets", ids, c);
        free(ids);
    }

    if (whupsey_get_requested_tickets(w, &ids, &c) == WHUPSEY_RPCFAULT)
    {
        printf("Error: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    }
    else
    {
        display_ids(w, "Requested Tickets", ids, c);
        free(ids);
    }
}

void test_create_ticket(Whupsey *w)
{
    WhupseyTicket ticket;

    const WhupseyModule  *modules = NULL;
    const WhupseyVersion *version = NULL;
    const WhupseyState   *state   = NULL;
    const WhupseyType    *types   = NULL;

    int nmodules;
    int ntypes;
    int count;
    int i;

    printf("----------------------------\n");
    printf("Attempting to create ticket:\n");

    whupsey_get_modules (w, &modules, &nmodules);
    whupsey_get_types   (w, &types,   &ntypes);

    if (nmodules == 0)
    {
        printf("  Can't create ticket - no modules available\n");
        return;
    }

    if (modules->ntypes == 0)
    {
        printf("  Can't create ticket - no types associated with the first module\n");
        return;
    }

    if (modules->types[0]->npriorities == 0)
    {
        printf("  Can't create ticket - no priorities associated with type %s\n", modules->types[0]->name);
        return;
    }

    state = modules[0].types[0]->states;
    count = modules[0].types[0]->nstates;
    for (i = 0; i < count; i++, state++)
        if (state->category == WHUPSEY_UNCONFIRMED)
            break;

    if (i == count)
    {
        state = modules[0].types[0]->states;
        count = modules[0].types[0]->nstates;
        for (i = 0; i < count; i++, state++)
            if (state->category == WHUPSEY_NEW)
                break;

        if (i == count)
        {
            printf("  Can't create ticket - no unconfirmed or new states in type %s\n", modules[0].types[0]->name);
            return;
        }
    }

    if (modules[0].has_versions)
    {
        if (modules[0].nversions == 0)
        {
            printf("  Can't create ticket - module %s is versioned but with no versions defined\n", modules[0].name);
            return;
        }

        version = modules[0].versions;
    }

    printf("  Module:   %s\n", modules[0].name);
    printf("  Type:     %s\n", modules[0].types[0]->name);
    printf("  State:    %s\n", state->name);
    printf("  Priority: %s\n", modules[0].types[0]->priorities[0].name);
    printf("  Version:  %s\n", version ? version->name : "(No versions)");
    printf("\n");

    ticket.id          = 0; // will be set by whupsey_add_ticket (if successful)
    ticket.comment_id  = 0; //   "
    ticket.timestamp   = 0; //   "

    ticket.type_id     = modules[0].types[0]->id;
    ticket.state_id    = state->id;
    ticket.priority_id = modules[0].types[0]->priorities[0].id;
    ticket.module_id   = modules[0].id;
    ticket.version_id  = 0;

    ticket.requester   = NULL;
    ticket.owner       = NULL;
    ticket.summary     = "RPC Generated Ticket";
    ticket.comment     = "This ticket was created by the WhupseyTest whups client.\n\nTesty test.\n";

    if (whupsey_add_ticket(w, &ticket) != WHUPSEY_OK)
        printf("Error: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    else
    {
        printf("New ticket id: %d, New comment id: %d, Timestamp %ld\n", ticket.id, ticket.comment_id, ticket.timestamp);

//        printf("\nTrying to set the attributes:\n");
//        test_set_attributes(w, ticket.id, modules[0].types[0]);
    }

    printf("\n");
}

void test_display_types(Whupsey *w)
{
    int ntypes;
    const WhupseyType *types;

    int i, j;

    whupsey_get_types(w, &types, &ntypes);

    printf("Got %d types:\n", ntypes);
    for (i = 0; i < ntypes; i++)
    {
        printf("    Id:         %d\n", types[i].id);
        printf("    Name:       %s\n", types[i].name);
        printf("    Desc:       %s\n", types[i].description);
        printf("    States:     %d states for this type:", types[i].nstates);

        for (j = 0; j < types[i].nstates; j++)
        {
            printf("\n");
            printf("        Id:       %d\n", types[i].states[j].id);
            printf("        Name:     %s\n", types[i].states[j].name);
            printf("        Desc:     %s\n", types[i].states[j].description);
            printf("        Category: %s\n", statec[types[i].states[j].category]);
        }

        printf("\n    Priorities: %d priorities for this type:\n", types[i].npriorities);

        for (j = 0; j < types[i].npriorities; j++)
        {
            printf("        Id:       %d\n", types[i].priorities[j].id);
            printf("        Name:     %s\n", types[i].priorities[j].name);
            printf("        Desc:     %s\n", types[i].priorities[j].description);
            printf("\n");
        }

        if (!types[i].npriorities)
            printf("\n");

        printf("    Attributes: %d attributes for this type:\n", types[i].nattributes);

        for (j = 0; j < types[i].nattributes; j++)
        {
            printf("        Id:       %d\n", types[i].attributes[j].id);
            printf("        Name:     %s\n", types[i].attributes[j].name);
            printf("        Desc:     %s\n", types[i].attributes[j].description);
            printf("\n");
        }

        if (!types[i].nattributes)
            printf("\n");
    }

    printf("\n");
}

void test_display_modules(Whupsey *w)
{
    int nmodules;
    const WhupseyModule *modules;

    int i, j;

    whupsey_get_modules(w, &modules, &nmodules);

    printf("Got %d modules:\n", nmodules);
    for (i = 0; i < nmodules; i++)
    {
        printf("    Id:        %d\n", modules[i].id);
        printf("    Name:      %s\n", modules[i].name);
        printf("    Desc:      %s\n", modules[i].description);
        printf("    Versioned: %s\n", modules[i].has_versions ? "yes" : "no");
        printf("    Subjects:  %s\n", modules[i].has_subjects ? "yes" : "no");
        printf("    Types:     %d types for this module:\n", modules[i].ntypes);

        for (j = 0; j < modules[i].ntypes; j++)
            printf("        Type %d (%s)\n", modules[i].types[j]->id, modules[i].types[j]->name);

        if (modules[i].has_versions)
        {
            printf("    Versions:  %d versions for this module:\n", modules[i].nversions);

            for (j = 0; j < modules[i].ntypes; j++)
            {
                printf("        Id:   %d\n", modules[i].versions[j].id);
                printf("        Name: %s\n", modules[i].versions[j].name);
                printf("        Desc: %s\n", modules[i].versions[j].description);
                printf("\n");
            }
        }

        printf("\n");
    }
}

void test_set_attributes(Whupsey *w, unsigned int ticket_id, WhupseyType *type)
{
    // A bit hardcoded at the moment, so not run by default

    WhupseyAttribute attribs[2];

    attribs[0].descriptor = whupsey_find_attribute_desc_by_name(w, type, "Test Attribute 1");
    attribs[1].descriptor = whupsey_find_attribute_desc_by_name(w, type, "Test Attribute 2");

    if (!attribs[0].descriptor || !attribs[1].descriptor)
    {
        printf("test_set_attributes: can't find attribute descriptors\n");
        return;
    }

    attribs[0].value = "Test 1";
    attribs[1].value = "Wibbly wibbly woo";

    if (whupsey_set_ticket_attributes(w, ticket_id, "Mr. Comment\n", attribs, 2) != WHUPSEY_OK)
        printf("Error setting attributes: %d, %s\n", whupsey_get_fault_code(w), whupsey_get_fault_string(w));
    else
        printf("Successfully set the ticket attributes\n");
}

