/////////////////////////////////////////////////////////////////
//
//  Whupsey - A Whups Client Library in C
//
//  21.11.2002
//

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "whupsey.h"

static void whupsey_free_schema     (Whupsey *w);
static void whupsey_free_types      (Whupsey *w);
static void whupsey_free_states     (Whupsey *w, WhupseyState *states, int nstates);
static void whupsey_free_priorities (Whupsey *w, WhupseyPriority *priorities, int npriorities);
static void whupsey_free_modules    (Whupsey *w);

struct Whupsey
{
    xmlrpc_env          env;
    xmlrpc_server_info *server;

    int ntypes;
    WhupseyType *types;

    int nmodules;
    WhupseyModule *modules;
};

static inline void CleanWhupsey(Whupsey *w)
{
    if (w->env.fault_occurred)
    {
        xmlrpc_env_clean(&w->env);

        w->env.fault_occurred = 0;
        w->env.fault_code     = 0;
        w->env.fault_string   = NULL;
    }
}

static inline void DumpValue(Whupsey *w, xmlrpc_value *v)
{
    int s;
    xmlrpc_mem_block *b = xmlrpc_mem_block_new(&w->env, 0);
    xmlrpc_serialize_value(&w->env, b, v);
    s = xmlrpc_mem_block_size(b);
    printf("Result:\n%*.*s\n\n", s, s, (char *) xmlrpc_mem_block_contents(b));
    xmlrpc_mem_block_free(b);
}


// Simple accessor methods first

int whupsey_get_fault_code(Whupsey *w)
{
    return w->env.fault_code;
}

const char *whupsey_get_fault_string(Whupsey *w)
{
    return w->env.fault_string;
}

void whupsey_get_types(Whupsey *w, const WhupseyType **types, int *ntypes)
{
    *ntypes = w->ntypes;
    *types  = w->types;
}

void whupsey_get_modules(Whupsey *w, const WhupseyModule **modules, int *nmodules)
{
    *nmodules = w->nmodules;
    *modules  = w->modules;
}

const WhupseyType *whupsey_find_type_by_id(Whupsey *w, unsigned int id)
{
    int i;

    for (i = 0; i < w->ntypes; i++)
        if (w->types[i].id == id)
            return &w->types[i];

    return NULL;
}

const WhupseyType *whupsey_find_type_by_name(Whupsey *w, const char *name)
{
    int i;

    for (i = 0; i < w->ntypes; i++)
        if (!strcmp(w->types[i].name, name))
            return &w->types[i];

    return NULL;
}

const WhupseyAttributeDesc *whupsey_find_attribute_desc_by_id(Whupsey *w, WhupseyType *type, unsigned int id)
{
    int i;

    for (i = 0; i < type->nattributes; i++)
        if (type->attributes[i].id == id)
            return &type->attributes[i];

    return NULL;
}

const WhupseyAttributeDesc *whupsey_find_attribute_desc_by_name(Whupsey *w, WhupseyType *type, const char *name)
{
    int i;

    for (i = 0; i < type->nattributes; i++)
        if (!strcmp(type->attributes[i].name, name))
            return &type->attributes[i];

    return NULL;
}


// Now for the real stuff

WhupseyResult whupsey_create(Whupsey **result, const char *useragent, const char *version, const char *server, const char *user, const char *pass)
{
    Whupsey *w = calloc(1, sizeof(Whupsey));

    *result = NULL;

    if (w == NULL)
        return WHUPSEY_NOMEMORY;

    xmlrpc_client_init(XMLRPC_CLIENT_NO_FLAGS, (char *) useragent, (char *) version);
    xmlrpc_env_init(&w->env);

    w->server = xmlrpc_server_info_new(&w->env, (char *) server);
    if (!w->env.fault_occurred)
        xmlrpc_server_info_set_basic_auth(&w->env, w->server, (char *) user, (char *) pass);

    if (w->env.fault_occurred)
    {
        whupsey_free(w);
        return WHUPSEY_RPCFAULT;
    }

    *result = w;
    return WHUPSEY_OK;
}

static WhupseyResult whupsey_get_int_array(Whupsey *w, const char *method,  int **result, int *count)
{
    xmlrpc_value *ids;
    xmlrpc_value *id;
    int size, i;

    CleanWhupsey(w);

    *result = NULL;
    *count  = 0;

    ids = xmlrpc_client_call_server(&w->env, w->server, (char *) method, "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    size = xmlrpc_array_size(&w->env, ids);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    *result = malloc(sizeof(int) * size);
    if (!*result)
        return WHUPSEY_NOMEMORY;

    for (i = 0; i < size; i++)
    {
        id = xmlrpc_array_get_item(&w->env, ids, i);
        if (!w->env.fault_occurred)
            xmlrpc_parse_value(&w->env, id, "i", *result + i);

        if (w->env.fault_occurred)
        {
            xmlrpc_DECREF(ids);
            free(*result);
            *result = NULL;
            return WHUPSEY_RPCFAULT;
        }
    }

    xmlrpc_DECREF(ids);
    *count = size;
    return WHUPSEY_OK;
}

WhupseyResult whupsey_get_assigned_tickets(Whupsey *w, int **result, int *count)
{
    return whupsey_get_int_array(w, "projects.get_assigned_ticket_ids", result, count);
}

WhupseyResult whupsey_get_requested_tickets(Whupsey *w, int **result, int *count)
{
    return whupsey_get_int_array(w, "projects.get_requested_ticket_ids", result, count);
}

WhupseyResult whupsey_add_ticket(Whupsey *w, WhupseyTicket *t)
{
    xmlrpc_value *result;

    int  id;
    int  comment_id;
    long timestamp;

    CleanWhupsey(w);

    // simple (but incomplete) tests for validity - of course the server
    // checks all arguments aswell
    if (!w || !t || t->summary == NULL || t->comment == NULL)
        return WHUPSEY_BADARG;

    result = xmlrpc_client_call_server(
        &w->env, w->server, "projects.add_ticket",
        "({s:i,s:i,s:i,s:i,s:i,s:s,s:s,s:s})",
        "type",     t->type_id,
        "state",    t->state_id,
        "priority", t->priority_id,
        "module",   t->module_id,
        "version",  t->version_id,
        "owner",    t->owner ? t->owner : "",
        "summary",  t->summary,
        "comment",  t->comment);

    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    xmlrpc_parse_value(
        &w->env, result,
        "{s:i,s:i,s:i,*}",
        "id",         &id,
        "comment_id", &comment_id,
        "timestamp",  &timestamp);

    xmlrpc_DECREF(result);

    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    t->id         = id;
    t->comment_id = comment_id;
    t->timestamp  = timestamp;

    return WHUPSEY_OK;
}

WhupseyResult whupsey_set_ticket_attributes(Whupsey *w, unsigned int ticket_id, const char *comment, WhupseyAttribute *attribs, int nattribs)
{
    xmlrpc_value *attrib_array;
    xmlrpc_value *entry;
    xmlrpc_value *result;
    int comment_id;
    int i;

    CleanWhupsey(w);

    if (nattribs == 0)
        return WHUPSEY_OK;

    if (!w || !ticket_id || attribs == NULL)
        return WHUPSEY_BADARG;

    attrib_array = xmlrpc_build_value(&w->env, "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    for (i = 0; i < nattribs; i++)
    {
        entry = xmlrpc_build_value(&w->env,
                    "{s:i,s:s}",
                    "attribute_id",    attribs[i].descriptor->id,
                    "attribute_value", attribs[i].value);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_array_append_item(&w->env, attrib_array, entry);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;
    }

    result = xmlrpc_client_call_server(
        &w->env, w->server, "projects.set_ticket_attributes",
        "({s:i,s:s,s:V})",
        "ticket_id",  ticket_id,
        "comment",    comment,
        "attributes", attrib_array);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    xmlrpc_parse_value(&w->env, result, "i", &comment_id);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    return WHUPSEY_OK;
}


///////////////////// Schema loading

static WhupseyResult whupsey_parse_states(Whupsey *w, xmlrpc_value *states, WhupseyType *type)
{
    char *name;
    char *description;
    char *category;
    int size, i;

    size = xmlrpc_array_size(&w->env, states);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    type->states  = calloc(size, sizeof(WhupseyState));
    type->nstates = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *state = xmlrpc_array_get_item(&w->env, states, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, state,
            "{s:i,s:s,s:s,s:s,*}",
            "state_id",          &type->states[i].id,
            "state_name",        &name,
            "state_description", &description,
            "state_category",    &category);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;
        
        type->states[i].name        = strdup(name);
        type->states[i].description = strdup(description);

        if (!strcasecmp(category, "unconfirmed"))
            type->states[i].category = WHUPSEY_UNCONFIRMED;
        else if (!strcasecmp(category, "new"))
            type->states[i].category = WHUPSEY_NEW;
        else if (!strcasecmp(category, "assigned"))
            type->states[i].category = WHUPSEY_ASSIGNED;
        else if (!strcasecmp(category, "resolved"))
            type->states[i].category = WHUPSEY_RESOLVED;
        else
            // Return this error to the user somehow (set w->env.fault_code?)
            return WHUPSEY_RPCFAULT;
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_priorities(Whupsey *w, xmlrpc_value *priorities, WhupseyType *type)
{
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, priorities);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    type->priorities  = calloc(size, sizeof(WhupseyPriority));
    type->npriorities = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *priority = xmlrpc_array_get_item(&w->env, priorities, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, priority,
            "{s:i,s:s,s:s,*}",
            "priority_id",          &type->priorities[i].id,
            "priority_name",        &name,
            "priority_description", &description);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        type->priorities[i].name        = strdup(name);
        type->priorities[i].description = strdup(description);
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_attribute_descs(Whupsey *w, xmlrpc_value *attribs, WhupseyType *type)
{
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, attribs);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    type->attributes  = calloc(size, sizeof(WhupseyAttributeDesc));
    type->nattributes = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *attrib = xmlrpc_array_get_item(&w->env, attribs, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, attrib,
            "{s:i,s:s,s:s,*}",
            "attribute_id",          &type->attributes[i].id,
            "attribute_name",        &name,
            "attribute_description", &description);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        type->attributes[i].name        = strdup(name);
        type->attributes[i].description = strdup(description);
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_types(Whupsey *w, xmlrpc_value *types)
{
    WhupseyResult r;
    xmlrpc_value *states;
    xmlrpc_value *priorities;
    xmlrpc_value *attributes;
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, types);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    w->types = calloc(size, sizeof(WhupseyType));
    if (!w->types)
        return WHUPSEY_NOMEMORY;

    w->ntypes = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *type = xmlrpc_array_get_item(&w->env, types, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, type,
            "{s:i,s:s,s:s,s:A,s:A,s:A,*}",
            "type_id",          &w->types[i].id,
            "type_name",        &name,
            "type_description", &description,
            "type_states",      &states,
            "type_priorities",  &priorities,
            "type_attributes",  &attributes);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        w->types[i].name        = strdup(name);
        w->types[i].description = strdup(description);

        r = whupsey_parse_states(w, states, w->types + i);
        if (r != WHUPSEY_OK)
            return r;

        r = whupsey_parse_priorities(w, priorities, w->types + i);
        if (r != WHUPSEY_OK)
            return r;

        r = whupsey_parse_attribute_descs(w, attributes, w->types + i);
        if (r != WHUPSEY_OK)
            return r;
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_type_ids(Whupsey *w, xmlrpc_value *type_ids, WhupseyModule *m)
{
    unsigned int id;
    int size, i;

    size = xmlrpc_array_size(&w->env, type_ids);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    m->types = calloc(size, sizeof(WhupseyType *));
    if (!m->types)
        return WHUPSEY_NOMEMORY;

    m->ntypes = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *type = xmlrpc_array_get_item(&w->env, type_ids, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(&w->env, type, "i", &id);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        m->types[i] = (WhupseyType *) whupsey_find_type_by_id(w, id);
        if (m->types[i] == NULL)
            return WHUPSEY_RPCFAULT; // FIXME: Better way to return this error
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_versions(Whupsey *w, xmlrpc_value *versions, WhupseyModule *m)
{
    char *name;
    char *description;
    int size, i;

    size = xmlrpc_array_size(&w->env, versions);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    if (size == 0)
        return WHUPSEY_OK;

    m->versions = calloc(size, sizeof(WhupseyVersion));
    if (!m->versions)
        return WHUPSEY_NOMEMORY;

    m->nversions = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *version = xmlrpc_array_get_item(&w->env, versions, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, version,
            "{s:i,s:s,s:s,*}",
            "version_id",          &m->versions[i].id,
            "version_name",        &name,
            "version_description", &description);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        m->versions[i].name        = strdup(name);
        m->versions[i].description = strdup(description);
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_modules(Whupsey *w, xmlrpc_value *modules)
{
    WhupseyResult r;
    xmlrpc_value *versions;
    xmlrpc_value *type_ids;

    char *name;
    char *description;

    int size, i;

    size = xmlrpc_array_size(&w->env, modules);
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    w->modules = calloc(size, sizeof(WhupseyModule));
    if (!w->modules)
        return WHUPSEY_NOMEMORY;

    w->nmodules = size;

    for (i = 0; i < size; i++)
    {
        xmlrpc_value *module = xmlrpc_array_get_item(&w->env, modules, i);
        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        xmlrpc_parse_value(
            &w->env, module,
            "{s:i,s:s,s:s,s:i,s:i,s:A,s:A,*}",
            "module_id",          &w->modules[i].id,
            "module_name",        &name,
            "module_description", &description,
            "module_subjectlist", &w->modules[i].has_subjects,
            "module_versioned",   &w->modules[i].has_versions,
            "module_versions",    &versions,
            "module_type_ids",    &type_ids);

        if (w->env.fault_occurred)
            return WHUPSEY_RPCFAULT;

        w->modules[i].name        = strdup(name);
        w->modules[i].description = strdup(description);

        r = whupsey_parse_type_ids(w, type_ids, w->modules + i);
        if (r != WHUPSEY_OK)
            return r;

        r = whupsey_parse_versions(w, versions, w->modules + i);
        if (r != WHUPSEY_OK)
            return r;
    }

    return WHUPSEY_OK;
}

static WhupseyResult whupsey_parse_schema(Whupsey *w, xmlrpc_value *schema)
{
    WhupseyResult  r;
    xmlrpc_value  *types;
    xmlrpc_value  *modules;

    xmlrpc_parse_value(
        &w->env, schema,
        "{s:A,s:A,*}",
        "types",   &types,
        "modules", &modules);

    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    r = whupsey_parse_types(w, types);
    if (r != WHUPSEY_OK)
        return r;

    r = whupsey_parse_modules(w, modules);
    if (r != WHUPSEY_OK)
        return r;

    return WHUPSEY_OK;    
}

WhupseyResult whupsey_load_schema(Whupsey *w)
{
    xmlrpc_value  *schema = NULL;
    WhupseyResult  r;

    CleanWhupsey(w);

    if (!w)
        return WHUPSEY_BADARG;

    whupsey_free_schema(w);

    schema = xmlrpc_client_call_server(&w->env, w->server, "projects.get_schema", "()");
    if (w->env.fault_occurred)
        return WHUPSEY_RPCFAULT;

    r = whupsey_parse_schema(w, schema);
    if (r != WHUPSEY_OK)
        whupsey_free_schema(w);

    xmlrpc_DECREF(schema);
    return r;
}

void whupsey_free(Whupsey *w)
{
    whupsey_free_schema(w);
    xmlrpc_env_clean(&w->env);

    if (w->server)
        xmlrpc_server_info_free(w->server);

    xmlrpc_client_cleanup();
    free(w);
}

static void whupsey_free_schema(Whupsey *w)
{
    whupsey_free_types(w);
    whupsey_free_modules(w);
}

static void whupsey_free_types(Whupsey *w)
{
    int i;

    if (w->types != NULL)
    {
        for (i = 0; i < w->ntypes; i++)
        {
            free(w->types[i].name);
            free(w->types[i].description);
            whupsey_free_states(w, w->types[i].states, w->types[i].nstates);
            whupsey_free_priorities(w, w->types[i].priorities, w->types[i].npriorities);
        }

        free(w->types);
    }

    w->ntypes = 0;
    w->types  = NULL;
}

static void whupsey_free_states(Whupsey *w, WhupseyState *states, int nstates)
{
    int i;

    if (!states)
        return;

    for (i = 0; i < nstates; i++)
    {
        free(states[i].name);
        free(states[i].description);
    }

    free(states);
}

static void whupsey_free_priorities(Whupsey *w, WhupseyPriority *priorities, int npriorities)
{
    int i;

    if (!priorities)
        return;

    for (i = 0; i < npriorities; i++)
    {
        free(priorities[i].name);
        free(priorities[i].description);
    }

    free(priorities);
}

static void whupsey_free_versions(Whupsey *w, WhupseyVersion *versions, int nversions)
{
    int i;

    if (!versions)
        return;

    for (i = 0; i < nversions; i++)
    {
        free(versions[i].name);
        free(versions[i].description);
    }

    free(versions);
}

static void whupsey_free_modules(Whupsey *w)
{
    int i;

    if (w->modules != NULL)
    {
        for (i = 0; i < w->nmodules; i++)
        {
            free(w->modules[i].name);
            free(w->modules[i].description);
            free(w->modules[i].types);

            whupsey_free_versions(w, w->modules[i].versions, w->modules[i].nversions);
        }

        free(w->modules);
    }

    w->nmodules = 0;
    w->modules  = NULL;
}

